"""Within a flow the difference in archival rates between the series is investigated.

<p class="scoring-explanation">The score of this check is calculated based on the
minimum (f_min) and maximum (f_max) archival frequency.</p>

<div class="ts-check-impact">
<p>Differences in archival rates means that for downstream analysis many more interpolated
values might need to be used for at least on of the series. This could hinder reliability
of the outcome of those analysis.</p>
</div>
"""

import pandas as pd
import numpy as np

from timeseer import (
    AnalysisInput,
    AnalysisResult,
    MultivariateAnalysisInput,
    Statistic,
)


_STATISTIC_NAME = "Archival consistency"

META = {
    "run": "before",
    "statistics": [
        {"name": _STATISTIC_NAME},
    ],
    "conditions": [
        {
            "min_series": 2,
        }
    ],
    "signature": "multivariate",
}


def _consistency_score(archivals: pd.DataFrame):
    sorted_archivals = np.sort(archivals)
    sorted_archivals = np.round(sorted_archivals / 10) * 10
    log_diff = np.log10(np.nanmax(sorted_archivals) / np.nanmin(sorted_archivals))
    _, cts = np.unique(sorted_archivals, return_counts=True)
    pcts = cts / len(sorted_archivals)
    return max(pcts) / (log_diff + 1)


def _get_archival_time_median(statistics: list[Statistic]):
    median_archival_step = [
        statistic.result
        for statistic in statistics
        if statistic.name == "Archival time median"
    ]
    if median_archival_step is None or len(median_archival_step) == 0:
        return np.nan
    return median_archival_step[0]


def _run_archival_ratio(inputs: list[AnalysisInput]):
    archival_rates = pd.DataFrame(
        map(lambda x: _get_archival_time_median(x.statistics), inputs)
    )
    archival_rates = archival_rates.dropna()
    if len(archival_rates) == 0:
        return None
    return _consistency_score(archival_rates)


# pylint: disable=missing-function-docstring
def run(
    analysis_input: MultivariateAnalysisInput,
) -> AnalysisResult:
    inputs = analysis_input.inputs

    archival_ratio = _run_archival_ratio(inputs)
    if archival_ratio is None:
        return AnalysisResult()
    statistic = Statistic(_STATISTIC_NAME, "float", archival_ratio)
    return AnalysisResult(statistics=[statistic])



"""The data in a time series should be present within the analysis timeframe.

<p>
The data in a time series should be present within the analysis timeframe.
</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>
No input - No output.
</p>
</div>
"""

from timeseer import AnalysisInput, AnalysisResult, EventFrame

_CHECK_NAME = "Data available"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": ["No data"],
            "data_type": "bool",
        }
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    no_data = len(analysis_input.data) == 0
    frames = []
    if no_data:
        frames.append(
            EventFrame(
                type="No data",
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=analysis_input.evaluation_time_range.end_date,
            )
        )
    return AnalysisResult(event_frames=frames, condition_message="OK exclude")



"""Interpolation information loss over all series in a flow.

<p>The Interpolation Information Loss (IIL) is a measure to indicate how much
of the variance is lost when interpolating time series at a particular frequency.
It is well known that just resampling at fixed intervals can cause missing extremes
and hence create data that is no longer representative for the system.

The multivariate IIL assumes a worst case and aggregates all individual IILs accordingly</p>"""

import itertools

from typing import List

import timeseer

from timeseer import DataType


META = {
    "run": "before",
    "conditions": [
        {
            "min_series": 2,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "multivariate",
}

sort_mapping = {
    "1s": 1,
    "5s": 5,
    "15s": 15,
    "1m": 60,
    "5m": 300,
    "15m": 900,
    "1h": 3600,
    "12h": 3600 * 12,
    "24h": 3600 * 24,
}


def _by_time(name: str):
    return sort_mapping[name[32:-1]]


def _run_multivariate_interpolation_information_loss(
    analysis_input: timeseer.MultivariateAnalysisInput,
) -> List[timeseer.Statistic]:
    stats = list(
        itertools.chain.from_iterable(
            [info.statistics for info in analysis_input.inputs]
        )
    )

    names = list(
        set(
            stat.name for stat in stats if "Interpolation information loss" in stat.name
        )
    )

    statistics = []
    for name in names:
        results = [stat.result for stat in stats if stat.name == name]
        if len(results) != len(analysis_input.inputs):
            continue
        statistics.append(
            timeseer.Statistic(f"Multivariate {name.lower()}", "hidden", max(results))
        )

    table_names = list(
        set(
            stat.name
            for stat in stats
            if "Interpolation information loss (" in stat.name
        )
    )
    values = []
    for name in sorted(table_names, key=_by_time):
        results = [stat.result for stat in stats if stat.name == name]
        if len(results) != len(analysis_input.inputs):
            continue
        values.append((name[32:-1], max(results)))

    table_statistics = timeseer.Statistic(
        "Multivariate Interpolation Information Loss", "table", values
    )
    return statistics, table_statistics


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.MultivariateAnalysisInput,
) -> timeseer.AnalysisResult:
    statistics, table_statistics = _run_multivariate_interpolation_information_loss(
        analysis_input
    )
    statistics.append(table_statistics)
    return timeseer.AnalysisResult(statistics=statistics)



import sqlite3

from yoyo import step

__depends__ = {"046__update_block_evaluations"}


def migrate_block_evaluations(db: sqlite3.Connection):
    cursor = db.cursor()
    cursor.execute("pragma schema_version")
    (schema_version,) = cursor.fetchone()
    cursor.execute("pragma writable_schema = ON")
    cursor.execute(
        """
        update sqlite_master
        set sql = 'CREATE TABLE BlockEvaluations (
    id integer primary key autoincrement,
    block_id integer,
    flow_evaluation_id integer not null,
    type text not null,
    name text not null,
    configuration text not null,
    sort_order integer not null,

    foreign key (flow_evaluation_id) references FlowEvaluations(id) on delete cascade
)'
        where type = 'table' and name = 'BlockEvaluations'
    """
    )
    cursor.execute(f"pragma schema_version = {schema_version + 1}")
    cursor.execute("pragma writable_schema = OFF")


steps = [
    step(migrate_block_evaluations),
]



"""For certain units (e.g. pH) there is a limited range values can take.

<p>This check validates whether for a specific set of units, the historical values are
within the expected deterministic range.</p>
<p class="scoring-explanation">The score for this check is based on the amount
of time values are outside of the unit range.</p>
<div class="ts-check-impact">
<p>Mismatches between unit range and measured values are data quality issues.
</p>
</div>
"""

from enum import Enum, auto
from datetime import datetime

import numpy as np
import pandas as pd

from pandas.api.types import is_string_dtype

from pint import UndefinedUnitError

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.analysis.utils import (
    get_unit_registry,
    event_frames_from_dataframe,
    merge_intervals_and_open_event_frames,
    process_open_intervals,
)
from timeseer.metadata import fields

_CHECK_NAME = "Values outside unit range"
_EVENT_FRAME_NAME = "Values outside unit range"


META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


class LimitType(Enum):
    """Choose whether to use physical or functional limits."""

    PHYSICAL = auto()
    FUNCTIONAL = auto()


def _make_units_dir():
    ureg = get_unit_registry()
    units_zero_min = [
        "meter",
        "second",
        "ampere",
        "candela",
        "gram",
        "mole",
        "kelvin",
        "unit",
        "pH",
        "m2",
        "liter",
        "hertz",
        "kph",
        "galileo",
        "newton",
        "joule",
        "watt",
        "water",
        "pascal",
        "foot_pound",
        "poise",
        "stokes",
        "rhe",
        "particle",
        "molar",
        "katal",
        "clausius",
        "entropy_unit",
        "curie",
        "langley",
        "nit",
        "lumen",
        "lux",
        "a_u_intensity",
        "volt",
        "ohm",
        "siemens",
        "henry",
        "weber",
        "tesla",
        "bohr_magneton",
    ]
    dimensions_zero_min = []
    for unit in units_zero_min:
        dimension = ureg.get_dimensionality(unit)
        dimensions_zero_min.append(dimension)
    return dimensions_zero_min


def _get_intervals(outliers, df, event_type):
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    outlier_intervals["type"] = event_type
    return outlier_intervals


def _get_active_points(df: pd.DataFrame, min_value: float, max_value: float):
    return (df["value"] < min_value) | (df["value"] > max_value)


def _handle_open_intervals(df, intervals):
    intervals["open"] = intervals["end_date"] == df.index[-1]
    return intervals


def _get_dimensionality(unit):
    ureg = get_unit_registry()
    try:
        return ureg.get_dimensionality(unit)
    except (UndefinedUnitError, AttributeError, ValueError, KeyError):
        return "Unknown"


def _unit_dir(unit):
    input_dimension = _get_dimensionality(unit)
    return input_dimension


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_values_outside_unit_range_check(
    unit, analysis_input: AnalysisInput, open_event_frames: list[EventFrame]
) -> tuple[list[EventFrame], datetime]:
    df = analysis_input.data
    df = _clean_dataframe(df)

    analysis_start = df.index[0]
    if analysis_input.analysis_start is not None:
        analysis_start = analysis_input.analysis_start

    min_value = 0
    max_value = np.inf
    if _unit_dir(unit) == _get_dimensionality("pH"):
        max_value = 14

    active_points = _get_active_points(df, min_value, max_value)
    intervals = _get_intervals(active_points, df, _EVENT_FRAME_NAME)
    intervals = _handle_open_intervals(df, intervals)

    intervals = merge_intervals_and_open_event_frames(
        analysis_start, intervals, open_event_frames
    )

    frames = event_frames_from_dataframe(process_open_intervals(intervals))

    last_analyzed_point = df.index[-1]

    return list(frames), last_analyzed_point


def _is_relevant_open_event_frame(event_frame):
    return (
        event_frame.end_date is None
        and event_frame.type in META["checks"][0]["event_frames"]
    )


def _get_open_event_frames(
    analysis_input: AnalysisInput,
) -> list[EventFrame]:
    return [
        frame
        for frame in analysis_input.event_frames
        if _is_relevant_open_event_frame(frame)
    ]


def _is_valid_input(
    analysis_input: AnalysisInput, median_archival_step: list[float], unit: str
) -> tuple[str, bool]:
    if median_archival_step is None or len(median_archival_step) == 0:
        return "No median archival step", False
    if len(_clean_dataframe(analysis_input.data)) == 0:
        return "No data", False
    if is_string_dtype(analysis_input.data["value"]):
        return "Can not be a string", False
    if unit is None:
        return "No unit", False
    dimensions_zero_min = _make_units_dir()
    if _unit_dir(unit) not in dimensions_zero_min:
        return "Unit not in dimensions", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    unit = analysis_input.metadata.get_field(fields.Unit)
    median_archival_step = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == "Archival time median"
    ]

    message, is_ok = _is_valid_input(analysis_input, median_archival_step, unit)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    open_event_frames = _get_open_event_frames(analysis_input)

    frames, last_analyzed_point = _run_values_outside_unit_range_check(
        unit, analysis_input, open_event_frames
    )
    return AnalysisResult(
        event_frames=frames,
        last_analyzed_point=last_analyzed_point,
    )



"""Calculate the histogram of a (unevenly spaced) time series."""

import numpy as np
import pandas as pd

import timeseer

from timeseer import DataType

META: dict = {
    "statistics": [{"name": "Archival Histogram"}],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _get_weights_unequally_spaced(df: pd.DataFrame):
    times = (
        (df.index.to_series() - df.index.to_series().shift()).dt.total_seconds()
    ).to_numpy()
    total_duration = np.nansum(times)
    weights = np.empty(len(times))
    weights[0] = 0.5 * times[1] / total_duration
    weights[-1] = 0.5 * times[-1] / total_duration
    if len(times) > 2:
        for i in range(1, (len(times) - 1)):
            weights[i] = 0.5 * (times[i] + times[i + 1]) / total_duration
    return weights


def _remove_outliers(values, weights):
    q25, q75 = np.nanquantile(values, [0.25, 0.75])
    iqr = q75 - q25
    return (
        values[(values <= q75 + 1.5 * iqr) & (values >= q25 - 1.5 * iqr)],
        weights[(values <= q75 + 1.5 * iqr) & (values >= q25 - 1.5 * iqr)],
    )


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].sort_index()


def _run_ts_archival_histogram(df: pd.DataFrame):
    clean_df = _clean_dataframe(df)
    if len(clean_df) < 30:
        return None, None

    meas_times = clean_df.index.to_series()
    diff_times = (meas_times - meas_times.shift()).dt.total_seconds()

    try:
        hist, bin_edges = np.histogram(
            diff_times,
            range=(np.nanmin(diff_times), np.nanmax(diff_times)),
            bins=20,
        )
        return hist, bin_edges
    except (ValueError, TypeError):
        return None, None


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    hist, bin_edges = _run_ts_archival_histogram(analysis_input.data)
    if hist is None:
        return timeseer.AnalysisResult(condition_message="No histogram")

    histogram = dict(hist=hist.tolist(), bin_edges=bin_edges.tolist())

    return timeseer.AnalysisResult(
        statistics=[
            timeseer.Statistic(META["statistics"][0]["name"], "histogram", histogram),
        ]
    )



"""Analysis on the variation in the archival time distribution per value range.

<p>Overcompression is a consequence of settings in a compression algorithm that
cause the archived data to no longer represent the real system accurately. Within traditional
historians this is typically caused by setting too high values for exception and/or compression.
A significant difference in median inter-archival time between normal values and extreme could indicate
issues with the compression parameters.</p>
<p><img src='../static/images/reporting/value_archival.svg'></p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>Badly compressed data, specifically overcompression,
can lead to critical events such as upsets, safety issues and downtime.
</p>
</div>
"""

import jsonpickle

import timeseer
from timeseer import DataType

_CHECK_NAME = "Compression - value dependent archival rate"

META = {
    "checks": [{"name": _CHECK_NAME, "data_type": "bool"}],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _run_value_dependent_archival_check(extremes_sketch, base_sketch):
    extremes_q25, extremes_q75 = [
        extremes_sketch.get_quantile_value(q) for q in [0.25, 0.75]
    ]
    base_q25, base_q75 = [base_sketch.get_quantile_value(q) for q in [0.25, 0.75]]
    if not (base_q25 > extremes_q75 or extremes_q25 > base_q75):
        return False
    return True


def _get_relevant_statistic(analysis_input, stat_name):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return jsonpickle.decode(statistics[0])


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.AnalysisInput,
) -> timeseer.AnalysisResult:
    extremes_sketch = _get_relevant_statistic(
        analysis_input, "Extremes Archival Sketch"
    )
    base_sketch = _get_relevant_statistic(analysis_input, "Base Archival Sketch")
    if extremes_sketch is None:
        return timeseer.AnalysisResult(condition_message="No extreme archival sketch")
    if base_sketch is None:
        return timeseer.AnalysisResult(condition_message="No base archival sketch")

    score = _run_value_dependent_archival_check(extremes_sketch, base_sketch)
    check = timeseer.CheckResult(_CHECK_NAME, float(score))
    return timeseer.AnalysisResult([check])



"""Check for a change after a critical event based on classification.

<p>This check identifies whether there is a significant difference in
a series before and after a critical event based on cluster distances.</p>
<div class="ts-check-impact">
<p>Changes in the behavior of a series after a critical event, such as a
maintenance, turnaround or other types of interventions could indicate
human error. This can range from physical damage to a sensor, to forgotten
control strategy updates.</p>
</div>"""

from typing import Union

import numpy as np
import pandas as pd

from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC

from timeseer import (
    AnalysisResult,
    CheckResult,
    ComparisonAnalysisInput,
    DataType,
    ProcessType,
)

_CHECK_NAME = "Comparison change classification"

META: dict = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "data_type": "bool",
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
            "process_type": [ProcessType.CONTINUOUS, ProcessType.REGIME],
        }
    ],
    "signature": "comparison",
}


def _clean_data(df: pd.DataFrame) -> pd.DataFrame:
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_compare_classication(
    analysis_input: ComparisonAnalysisInput,
) -> Union[bool, None]:
    df_before = _clean_data(analysis_input.reference_data)[["value"]]
    df_after = _clean_data(analysis_input.data)[["value"]]

    if len(df_before) == 0 or len(df_after) == 0:
        return None

    sel_before = np.random.choice(range(len(df_before)), 1000)
    sel_after = np.random.choice(range(len(df_after)), 1000)
    df_before = df_before.assign(classlabel=np.ones(len(df_before))).iloc[sel_before]
    df_after = df_after.assign(classlabel=np.zeros(len(df_after))).iloc[sel_after]
    full_df = pd.concat([df_before, df_after])
    X = full_df.iloc[:, 0:-1]
    y = full_df.iloc[:, -1]
    clf = make_pipeline(StandardScaler(), SVC(gamma="auto"))
    clf.fit(X, y)
    y_hat = clf.predict(X)
    score = sum(y == y_hat) / len(full_df)
    return score <= 0.8


# pylint: disable=missing-function-docstring
def run(
    analysis_input: ComparisonAnalysisInput,
) -> AnalysisResult:

    is_change = _run_compare_classication(analysis_input)
    if is_change is None:
        return AnalysisResult()

    check = CheckResult(_CHECK_NAME, float(is_change))

    return AnalysisResult(check_results=[check])



"""The series should have units.

<p>Units aid in the interpretation of measurement readings. Often series naming conventions
show some insight into the type of sensor (e.g. PI for pressure indicator).</p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).
This check belongs to the score group Unit.
If this check is True, it means that the score for group Unit will be 0%.</p>
<div class="ts-check-impact">
<p>
Interpreting a dimensionless variable can lead to wrong conclusions and actions
(e.g. bara vs barg vs kPa vs ...). When combining data from different sensors of a similar type,
comparisons without units are impossible. Without units the interpretation of the sensibility of
a measurement value is hindered.
</p>
</div>
"""

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.metadata import fields

_CHECK_NAME = "Unit is present"
_EVENT_FRAME_NAME = "Unit is present"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "group": "Unit - Physical",
            "event_frames": [_EVENT_FRAME_NAME],
            "data_type": "bool",
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    unit = analysis_input.metadata.get_field(fields.Unit)

    event_frames = []
    if unit == "":
        event_frames.append(
            EventFrame(
                type=_EVENT_FRAME_NAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=analysis_input.evaluation_time_range.end_date,
            )
        )
    return AnalysisResult(event_frames=event_frames)



import sqlite3

from yoyo import step

__depends__ = {"015__add_tags_field"}


def add_foreign_keys(db: sqlite3.Connection):
    cursor = db.cursor()
    cursor.execute("pragma schema_version")
    (schema_version,) = cursor.fetchone()
    cursor.execute("pragma writable_schema = ON")
    cursor.execute(
        """
        update sqlite_master
        set sql = 'create table SeriesTags (
    id integer primary key autoincrement,
    series_id text not null,
    tag_key text not null,
    tag_value text not null,

    unique(series_id, tag_key, tag_value),
    foreign key (series_id) references SeriesNames(series_id) on delete cascade
)'
        where type = 'table' and name = 'SeriesTags'
    """
    )
    cursor.execute(f"pragma schema_version = {schema_version + 1}")
    cursor.execute("pragma writable_schema = OFF")


steps = [
    step(add_foreign_keys),
]



"""Check for a change after a critical event based on classification.

<p>This check identifies whether there is a significant difference in
a series before and after a critical event based on cluster distances.</p>
<div class="ts-check-impact">
<p>Changes in the behavior of a series after a critical event, such as a
maintenance, turnaround or other types of interventions could indicate
human error. This can range from physical damage to a sensor, to forgotten
control strategy updates.</p>
</div>"""

import numpy as np
import pandas as pd

from scipy import stats

from timeseer import (
    AnalysisResult,
    CheckResult,
    ComparisonAnalysisInput,
    DataType,
    ProcessType,
)

_CHECK_NAME = "Comparison change batch length"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "data_type": "bool",
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
            "process_type": [ProcessType.BATCH],
        }
    ],
    "signature": "comparison",
}


def _clean_data(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _get_interval_lengths(outliers, df):
    outliers = pd.Series(data=outliers, index=df.index).fillna(False)
    outlier_grp = (outliers != outliers.shift().bfill()).cumsum()
    outlier_intervals = (
        df.assign(outlier_grp=outlier_grp)[outliers]
        .reset_index()
        .groupby(["outlier_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    interval_lengths = outlier_intervals["end_date"] - outlier_intervals["start_date"]
    return interval_lengths.dt.total_seconds()


def _run_compare_batch_length(analysis_input: ComparisonAnalysisInput):
    df_before = _clean_data(analysis_input.reference_data)
    df_after = _clean_data(analysis_input.data)

    if len(df_before) == 0 or len(df_after) == 0:
        return None

    nonzero_before = df_before["value"] > np.nanquantile(df_before["value"], 0.1)
    nonzero_after = df_after["value"] > np.nanquantile(df_after["value"], 0.1)
    before_lengths = _get_interval_lengths(nonzero_before, df_before)
    after_lengths = _get_interval_lengths(nonzero_after, df_after)

    if len(before_lengths) < 10 or len(after_lengths) < 10:
        return None

    if all(np.diff(np.concatenate((before_lengths, after_lengths))) == 0.0):
        return 1

    _, p_value = stats.ttest_ind(before_lengths, after_lengths)
    if np.isnan(p_value):
        return None

    return p_value >= 0.05


def run(
    analysis_input: ComparisonAnalysisInput,
):  # pylint: disable=missing-function-docstring

    is_change = _run_compare_batch_length(analysis_input)
    if is_change is None:
        return AnalysisResult()

    check = CheckResult(_CHECK_NAME, int(is_change))

    return AnalysisResult(check_results=[check])



""" Calculates the number of data points and the data points per day and per hour in a time series. """

from datetime import timedelta
from timeseer import AnalysisInput, AnalysisResult, Statistic

_STATISTIC_NAME = "Data points"

META = {
    "statistics": [
        {"name": _STATISTIC_NAME},
    ],
    "run": "before",
    "conditions": [
        {
            "min_series": 1,
        }
    ],
    "signature": "univariate",
}


def _get_timespan(analysis_input: AnalysisInput) -> timedelta:
    first_timestamp = analysis_input.data.index.to_series()[0].to_pydatetime()
    last_timestamp = analysis_input.data.index.to_series()[-1].to_pydatetime()
    return last_timestamp - first_timestamp


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    total_data_points = len(analysis_input.data.dropna())
    result = [("Total", total_data_points)]

    if total_data_points == 0:
        return AnalysisResult(statistics=[Statistic(_STATISTIC_NAME, "table", result)])

    timespan = _get_timespan(analysis_input)

    if timespan.days > 0:
        result.append(
            (
                "Average per day",
                int(total_data_points / (timespan.total_seconds() / 60 / 60 / 24)),
            )
        )

    if timespan.total_seconds() > 0:
        result.append(
            (
                "Average per hour",
                int(total_data_points / (timespan.total_seconds() / 60 / 60)),
            )
        )

    return AnalysisResult(statistics=[Statistic(_STATISTIC_NAME, "table", result)])



"""Time series should indicate their interpolation type.

<p>
It is often required to know the value of a time series between recorded data
points.
</p>
<p>
Two types of interpolation are commonly used:
</p>
<dl>
    <dt>LINEAR interpolation</dt>
    <dd>The value is interpolated linearly between the previously recorded and the next value.</dd>

    <dt>STEPPED interpolation</dt>
    <dd>The value is assumed to be equal to the previously recorded value.</dd>
</dl>
<p>
<p class="scoring-explanation">The score for this check is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>
Without a well-defined interpolation type, linear interpolation can happen when
stepped interpolation was required, for example for setpoints. Furthermore correlation analysis
between two series will be skewed when (at least one of) the series are not interpolated correct.
</p>
</div>
"""

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.metadata import fields

_CHECK_NAME = "Interpolation type is present"
_EVENT_FRAME_NAME = "Interpolation type is present"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "group": "Interpolation type",
            "event_frames": [_EVENT_FRAME_NAME],
            "data_type": "bool",
        }
    ],
    "conditions": [
        {"min_series": 1, "data_type": [DataType.FLOAT32, DataType.FLOAT64]}
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    interpolation_type = analysis_input.metadata.get_field(fields.InterpolationType)

    event_frames = []
    if interpolation_type is None:
        event_frames.append(
            EventFrame(
                type=_EVENT_FRAME_NAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=analysis_input.evaluation_time_range.end_date,
            )
        )

    return AnalysisResult(event_frames=event_frames)



"""All values should appear in the dictionary.

<p>
Dictionaries map numerical values to textual labels.
For example, a valve could be 'OPEN' (1) or 'CLOSED' (2).
If the numerical value 3 appears in the data stream, this value can not be
presented to analytics users and analytics tooling can fail because it can not
map the value.
</p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where a value is present that is not in the dictionary.
Imagine that 100 points are analyzed in a given time-frame and there are 10 points with values
outside the dictionary.
The score for this check in that case would be
90% = 1 - 10 / 100. Which means that for 90% of all points their value is in the dictionary.</p>
<div class="ts-check-impact">
<p>
Values outside the domain of the dictionary can hamper (autoamated) downstream analytics or visualization.
</p>
</div>
"""

from timeseer import AnalysisInput, AnalysisResult, DataType
from timeseer.analysis.utils import event_frames_from_dataframe
from timeseer.metadata import fields

_CHECK_NAME = "Values in dictionary"
_EVENT_FRAME_NAME = "Values not in dictionary"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": [_EVENT_FRAME_NAME],
        }
    ],
    "conditions": [
        {"min_series": 1, "min_data_points": 1, "data_type": [DataType.DICTIONARY]}
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(analysis_input: AnalysisInput) -> AnalysisResult:
    dictionary = analysis_input.metadata.get_field(fields.Dictionary)
    if dictionary is None or len(dictionary.mapping) == 0:
        return AnalysisResult(condition_message="No dictionary")
    data = analysis_input.data
    mask = data["value"].isin(dictionary.mapping.keys())

    interval_grp = (mask != mask.shift().bfill()).cumsum()
    intervals = (
        data.assign(interval_grp=interval_grp)[~mask]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = _EVENT_FRAME_NAME

    return AnalysisResult(
        event_frames=list(event_frames_from_dataframe(intervals)),
    )



"""Identification of outlier based on inter-quartile range (IQR).

<p>This check is a traditional method for outliers where values are flagged
as anomalies if they are outside of a range defined by [q1 - 1.5 * iqr, q3 + 1.5 * iqr].
Here q1, q3 and iqr stand for 25th, 75th percentile and inter-quartile range (=q3-q1) respectively.</p>
<p><img src='../static/images/reporting/iqr.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where an IQR outlier occurs. Imagine that 100 points are analyzed in a given time-frame
and there are 7 outliers. The score for this check in that case would be
93% = 1 - 7 / 100. Which means that 93% of all points are not outliers.</p>
<div class="ts-check-impact">
<p>
Outliers in series can cloud downstream analytics and put the burden of removing these without
background knowledge on the algorithms instead of those in the know.
Several calculations are sensitive to these type of sudden changes and process decision chains
based on such an outlier can propagate through the system.
Any type of anomaly detection based on normal behavior of the phyiscal sensor can act as a guide for
prioritization of callibration / maintenance.
</p>
</div>
"""

from datetime import datetime, timedelta

import jsonpickle
import pandas as pd

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.analysis.utils import (
    event_frames_from_dataframe,
    merge_intervals_and_open_event_frames,
    process_open_intervals,
)

_CHECK_NAME = "IQR Anomaly"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "event_frames": ["IQR Anomaly"],
        }
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


def _calculate_total_check_active_time(frames, median_archival_step):
    active_time = timedelta(0)
    for _, row in frames.iterrows():
        length = row["end_date"] - row["start_date"]
        if length == timedelta(0):
            length = timedelta(seconds=median_archival_step)
        active_time = active_time + length
    return active_time


def _handle_open_intervals(df, intervals):
    intervals["open"] = intervals["end_date"] == df.index[-1]
    return intervals


def _get_intervals(anomalies, df, event_type):
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()
    intervals = (
        df.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = event_type
    return intervals


def _clean_dataframe(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _run_iqr_ad(
    analysis_input: AnalysisInput,
    open_event_frames: list[EventFrame],
    sketch,
    strictness=3,
) -> tuple[list[EventFrame], datetime]:
    df = _clean_dataframe(analysis_input.data)

    analysis_start = df.index[0]
    if analysis_input.analysis_start is not None:
        analysis_start = analysis_input.analysis_start

    q25, q75 = [sketch.get_quantile_value(q) for q in [0.25, 0.75]]
    iqr = q75 - q25
    active_points = (df["value"] < (q25 - (strictness * iqr))) | (
        df["value"] > (q75 + (strictness * iqr))
    )

    intervals = _get_intervals(active_points, df, _CHECK_NAME)
    intervals = _handle_open_intervals(df, intervals)
    intervals = merge_intervals_and_open_event_frames(
        analysis_start, intervals, open_event_frames
    )

    all_frames = event_frames_from_dataframe(process_open_intervals(intervals))

    last_analyzed_point = df.index[-1]

    return list(all_frames), last_analyzed_point


def _get_relevant_statistic(analysis_input, stat_name):
    statistics = [
        statistic.result
        for statistic in analysis_input.statistics
        if statistic.name == stat_name
    ]
    if statistics is None or len(statistics) == 0:
        return None
    return statistics[0]


def _is_relevant_open_event_frame(event_frame):
    return (
        event_frame.end_date is None
        and event_frame.type in META["checks"][0]["event_frames"]
    )


def _get_open_event_frames(
    analysis_input: AnalysisInput,
) -> list[EventFrame]:
    return [
        frame
        for frame in analysis_input.event_frames
        if _is_relevant_open_event_frame(frame)
    ]


def _is_valid_input(data: pd.DataFrame) -> tuple[str, bool]:
    if len(_clean_dataframe(data)) == 0:
        return "No clean data", False
    return "OK", True


# pylint: disable=missing-function-docstring
def run(analysis_input: AnalysisInput) -> AnalysisResult:
    message, is_ok = _is_valid_input(analysis_input.data)
    if not is_ok:
        return AnalysisResult(condition_message=message)

    median_archival_step = _get_relevant_statistic(
        analysis_input, "Archival time median"
    )
    json_sketch = _get_relevant_statistic(analysis_input, "Value Sketch")
    if median_archival_step is None:
        return AnalysisResult(condition_message="No median archival step")
    if json_sketch is None:
        return AnalysisResult(condition_message="No value sketch")
    sketch = jsonpickle.decode(json_sketch)

    open_event_frames = _get_open_event_frames(analysis_input)

    frames, last_analyzed_point = _run_iqr_ad(analysis_input, open_event_frames, sketch)
    return AnalysisResult(event_frames=frames, last_analyzed_point=last_analyzed_point)



"""Check for a change after a critical event based on classification.

<p>This check identifies whether there is a significant difference in
a series before and after a critical event based on cluster distances.</p>
<div class="ts-check-impact">
<p>Changes in the behavior of a series after a critical event, such as a
maintenance, turnaround or other types of interventions could indicate
human error. This can range from physical damage to a sensor, to forgotten
control strategy updates.</p>
</div>"""


import pandas as pd
import numpy as np

from scipy.stats import chisquare

from timeseer import AnalysisResult, CheckResult, ComparisonAnalysisInput, DataType

_CHECK_NAME = "Comparison change digital states times"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "data_type": "bool",
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 1,
            "data_type": [DataType.STRING, DataType.DICTIONARY, DataType.CATEGORICAL],
        }
    ],
    "signature": "comparison",
}


def _clean_data(df: pd.DataFrame):
    return df[~df.index.duplicated(keep="first")].dropna().sort_index()


def _are_similar_state_times(before, after):
    before_values = np.array(before["state_times"])
    after_values = np.array(
        after["state_times"] / sum(after["state_times"]) * sum(before["state_times"])
    )
    _, p_value = chisquare(before_values, after_values)
    return p_value >= 0.05


def _run_compare_digital_states(analysis_input: ComparisonAnalysisInput):
    df_before = _clean_data(analysis_input.reference_data)
    df_after = _clean_data(analysis_input.data)

    if len(df_before) == 0 or len(df_after) == 0:
        return None

    if set(df_before["value"]) != set(df_after["value"]):
        return False

    df_before["state_times"] = list(
        (
            pd.Series(df_before.index).shift(-1) - pd.Series(df_before.index)
        ).dt.total_seconds()
    )
    df_after["state_times"] = list(
        (
            pd.Series(df_after.index).shift(-1) - pd.Series(df_after.index)
        ).dt.total_seconds()
    )

    return _are_similar_state_times(
        df_before.groupby(["value"]).sum(), df_after.groupby(["value"]).sum()
    )


def run(
    analysis_input: ComparisonAnalysisInput,
):  # pylint: disable=missing-function-docstring

    is_change = _run_compare_digital_states(analysis_input)
    if is_change is None:
        return AnalysisResult()

    check = CheckResult(_CHECK_NAME, float(is_change))

    return AnalysisResult(check_results=[check])



"""For certain units (e.g. pH) there is a limited range values can take.

<p>This check validates whether for a specific set of units, the configured limits fall
within the expected deterministic range.</p>
<p class="scoring-explanation">The score for this checks is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>Mismatches between unit range and physical limits are data hygiene bugs.
</p>
</div>
"""

from enum import Enum, auto
from typing import Optional

import numpy as np

from pint import UndefinedUnitError

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.analysis.utils import get_unit_registry
from timeseer.metadata import fields

_FUNCTIONAL_CHECK_NAME = "Limits inconsistent with unit (functional)"
_PHYSICAL_CHECK_NAME = "Limits inconsistent with unit (physical)"
_FUNCTIONAL_EVENT_FRAME_NAME = "Limits inconsistent with unit (functional)"
_PHYSICAL_EVENT_FRAME_NAME = "Limits inconsistent with unit (physical)"

META = {
    "checks": [
        {
            "name": _FUNCTIONAL_CHECK_NAME,
            "group": "Unit - Functional",
            "event_frames": [_FUNCTIONAL_EVENT_FRAME_NAME],
            "data_type": "bool",
        },
        {
            "name": _PHYSICAL_CHECK_NAME,
            "group": "Unit - Physical",
            "event_frames": [_PHYSICAL_EVENT_FRAME_NAME],
            "data_type": "bool",
        },
    ],
    "conditions": [
        {
            "min_series": 1,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [DataType.FLOAT32, DataType.FLOAT64],
        }
    ],
    "signature": "univariate",
}


class LimitType(Enum):
    """Choose whether to use physical or functional limits."""

    PHYSICAL = auto()
    FUNCTIONAL = auto()


def _make_units_dir():
    ureg = get_unit_registry()
    units_zero_min = [
        "meter",
        "second",
        "ampere",
        "candela",
        "gram",
        "mole",
        "kelvin",
        "unit",
        "pH",
        "m2",
        "liter",
        "hertz",
        "kph",
        "galileo",
        "newton",
        "joule",
        "watt",
        "water",
        "pascal",
        "foot_pound",
        "poise",
        "stokes",
        "rhe",
        "particle",
        "molar",
        "katal",
        "clausius",
        "entropy_unit",
        "curie",
        "langley",
        "nit",
        "lumen",
        "lux",
        "a_u_intensity",
        "volt",
        "ohm",
        "siemens",
        "henry",
        "weber",
        "tesla",
        "bohr_magneton",
    ]
    dimensions_zero_min = []
    for unit in units_zero_min:
        dimension = ureg.get_dimensionality(unit)
        dimensions_zero_min.append(dimension)
    return dimensions_zero_min


def _limit_settings_outside_range(
    analysis_input: AnalysisInput,
    limit_type: LimitType,
    min_value: float,
    max_value: float,
) -> Optional[bool]:
    if limit_type == LimitType.PHYSICAL:
        limit_low = analysis_input.metadata.get_field(fields.LimitLowPhysical)
        limit_high = analysis_input.metadata.get_field(fields.LimitHighPhysical)
    else:
        limit_low = analysis_input.metadata.get_field(fields.LimitLowFunctional)
        limit_high = analysis_input.metadata.get_field(fields.LimitHighFunctional)
    if limit_low and limit_low < min_value:
        return True
    if limit_high and limit_high > max_value:
        return True
    if not (limit_low or limit_high):
        return None
    return False


def _get_dimensionality(unit):
    ureg = get_unit_registry()
    try:
        return ureg.get_dimensionality(unit)
    except (UndefinedUnitError, AttributeError, ValueError, KeyError):
        return "Unknown"


def _unit_dir(unit):
    input_dimension = _get_dimensionality(unit)
    return input_dimension


def _run_values_outside_unit_range_check(
    unit, analysis_input: AnalysisInput, limit_type: LimitType
) -> Optional[bool]:
    dimensions_zero_min = _make_units_dir()
    if _unit_dir(unit) not in dimensions_zero_min:
        return None
    if _unit_dir(unit) == _get_dimensionality("pH"):
        return _limit_settings_outside_range(
            analysis_input, limit_type, min_value=0, max_value=14
        )
    return _limit_settings_outside_range(
        analysis_input, limit_type, min_value=0, max_value=np.inf
    )


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    unit = analysis_input.metadata.get_field(fields.Unit)
    if unit is None:
        return AnalysisResult(condition_message="No unit")

    event_frames = []

    functional_score = _run_values_outside_unit_range_check(
        unit, analysis_input, LimitType.FUNCTIONAL
    )
    if functional_score is not None and functional_score is True:
        event_frames.append(
            EventFrame(
                type=_FUNCTIONAL_EVENT_FRAME_NAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=analysis_input.evaluation_time_range.end_date,
            )
        )

    physical_score = _run_values_outside_unit_range_check(
        unit, analysis_input, LimitType.PHYSICAL
    )

    if physical_score is not None and physical_score is True:
        event_frames.append(
            EventFrame(
                type=_PHYSICAL_EVENT_FRAME_NAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=analysis_input.evaluation_time_range.end_date,
            )
        )

    return AnalysisResult(event_frames=event_frames)



"""For percentage accuracy the value should be between 0 and 100.

<p>This check validates whether the accuracy percentage falls
within the expected range between 0 and 100.</p>
<p class="scoring-explanation">The score for this checks is a simple boolean (True / False).</p>
<div class="ts-check-impact">
<p>Badly configured accuracy percentage can cause a wrong accuracy calculation.
</p>
</div>
"""

from timeseer import AnalysisInput, AnalysisResult, DataType, EventFrame
from timeseer.metadata import fields

_CHECK_NAME = "Accuracy percentage is within range"
_EVENT_FRAME_NAME = "Accuracy percentage is within range"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "group": "Accuracy",
            "event_frames": [_EVENT_FRAME_NAME],
            "data_type": "bool",
        }
    ],
    "conditions": [
        {"min_series": 1, "data_type": [DataType.FLOAT32, DataType.FLOAT64]}
    ],
    "signature": "univariate",
}


# pylint: disable=missing-function-docstring
def run(
    analysis_input: AnalysisInput,
) -> AnalysisResult:
    accuracy_percentage = analysis_input.metadata.get_field(fields.AccuracyPercentage)
    if accuracy_percentage is None:
        return AnalysisResult(condition_message="No accuracy percentage")
    event_frames = []
    if accuracy_percentage < 0 or accuracy_percentage > 100:
        event_frames.append(
            EventFrame(
                type=_EVENT_FRAME_NAME,
                start_date=analysis_input.evaluation_time_range.start_date,
                end_date=analysis_input.evaluation_time_range.end_date,
            )
        )
    return AnalysisResult(event_frames=event_frames)



"""Identify highly correlated tag pairs within a given set of tags.

<p>This returns suggestions for pairs to be treated as redundant sensors.</p>"""

from typing import List
from uuid import UUID

import networkx as nx
import numpy as np
import pandas as pd

from networkx.algorithms.components import connected_components

import timeseer

from timeseer import AnalysisInput, AnalysisResult, DataType, SeriesId
from timeseer.metadata import fields


_CHECK_NAME = "Series set discovery"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "data_type": "bool",
        },
    ],
    "conditions": [
        {
            "min_series": 2,
            "min_weeks": 0,
            "min_data_points": 0,
            "data_type": [
                DataType.FLOAT32,
                DataType.FLOAT64,
                DataType.DICTIONARY,
                DataType.CATEGORICAL,
            ],
        }
    ],
}


def _is_series_relevant(analysis_input: AnalysisInput) -> bool:
    data_type = analysis_input.metadata.get_field(fields.DataType)
    if data_type is not None and data_type == DataType.STRING:
        return False
    if analysis_input.calculated_metadata is not None:
        calculated_data_type = analysis_input.calculated_metadata.get_field(
            fields.DataType
        )
        if calculated_data_type is not None and calculated_data_type == DataType.STRING:
            return False
    return True


def _get_relevant_series(inputs) -> List[AnalysisInput]:
    return [
        analysis_input
        for analysis_input in inputs
        if _is_series_relevant(analysis_input)
    ]


def _run_correlation_recommendation(inputs: List[AnalysisInput]) -> List[List[UUID]]:
    relevant_series = _get_relevant_series(inputs)
    concatenated_df = (
        pd.concat(
            [
                series.data["value"][
                    ~series.data["value"].index.duplicated(keep="first")
                ]
                for series in relevant_series
            ],
            axis=1,
            sort=False,
        )
        .interpolate("linear")
        .dropna()
    )

    relevant_series_ids = np.array(
        [analysis_input.metadata.series.series_id for analysis_input in relevant_series]
    )
    concatenated_df.columns = relevant_series_ids
    concatenated_df = concatenated_df[(concatenated_df != 0).all(1)]
    concatenated_df = concatenated_df.dropna()

    correlation_matrix = abs(concatenated_df.corr()) > 0.9
    square_distance_matrix = np.array(correlation_matrix)

    connected_graph = nx.from_numpy_matrix(
        square_distance_matrix, create_using=nx.MultiGraph
    )
    graph_components = list(connected_components(connected_graph))
    return [
        relevant_series_ids[np.array(list(group))]
        for group in graph_components
        if len(group) > 1
    ]


# pylint: disable=missing-function-docstring
def run(
    analysis_input: timeseer.MultivariateAnalysisInput,
) -> AnalysisResult:
    inputs = analysis_input.inputs

    suggestions = _run_correlation_recommendation(inputs)
    return AnalysisResult(
        series_sets=dict(
            correlation=[
                [SeriesId(suggestion) for suggestion in group] for group in suggestions
            ]
        )
    )
